# outline for a myst_nb project with sphinx
# build with: sphinx-build -nW --keep-going -b html . ./_build/html

# load extensions
extensions = ['myst_nb']

# specify project details
master_doc = 'index'
project = 'NiAS'

# basic build settings
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', '**.ipynb_checkpoints']
nitpicky = True

## myst_nb default settings

# Custom formats for reading notebook; suffix -> reader
# nb_custom_formats = {}

# Notebook level metadata key for config overrides
# nb_metadata_key = 'mystnb'

# Cell level metadata key for config overrides
# nb_cell_metadata_key = 'mystnb'

# Mapping of kernel name regex to replacement kernel name(applied before execution)
# nb_kernel_rgx_aliases = {}

# Regex that matches permitted values of eval expressions
# nb_eval_name_regex = '^[a-zA-Z_][a-zA-Z0-9_]*$'

# Execution mode for notebooks
# nb_execution_mode = 'auto'

# Path to folder for caching notebooks (default: <outdir>)
# nb_execution_cache_path = ''

# Exclude (POSIX) glob patterns for notebooks
# nb_execution_excludepatterns = ()

# Execution timeout (seconds)
# nb_execution_timeout = 30

# Use temporary folder for the execution current working directory
# nb_execution_in_temp = False

# Allow errors during execution
# nb_execution_allow_errors = False

# Raise an exception on failed execution, rather than emitting a warning
# nb_execution_raise_on_error = False

# Print traceback to stderr on execution error
# nb_execution_show_tb = False

# Merge stdout/stderr execution output streams
# nb_merge_streams = False

# The entry point for the execution output render class (in group `myst_nb.output_renderer`)
# nb_render_plugin = 'default'

# Remove code cell source
# nb_remove_code_source = False

# Remove code cell outputs
# nb_remove_code_outputs = False

# Prompt to expand hidden code cell {content|source|outputs}
# nb_code_prompt_show = 'Show code cell {type}'

# Prompt to collapse hidden code cell {content|source|outputs}
# nb_code_prompt_hide = 'Hide code cell {type}'

# Number code cell source lines
# nb_number_source_lines = False

# Overrides for the base render priority of mime types: list of (builder name, mime type, priority)
# nb_mime_priority_overrides = ()

# Behaviour for stderr output
# nb_output_stderr = 'show'

# Pygments lexer applied to stdout/stderr and text/plain outputs
# nb_render_text_lexer = 'myst-ansi'

# Pygments lexer applied to error/traceback outputs
# nb_render_error_lexer = 'ipythontb'

# Options for image outputs (class|alt|height|width|scale|align)
# nb_render_image_options = {}

# Options for figure outputs (classes|name|caption|caption_before)
# nb_render_figure_options = {}

# The format to use for text/markdown rendering
# nb_render_markdown_format = 'commonmark'



# This file is part of the NiAS project (https://github.com/nias-project).
# Copyright NiAS developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

from abc import abstractmethod
from numbers import Number

import numpy as np
from numpy.typing import NDArray
from scipy.sparse import issparse

from nias.base.vectorarrays import VectorArrayBase, VectorArrayImpl
from nias.interfaces import (
    ArrayLike,
    Indices,
    VectorArray,
    VectorSpace,
    VectorSpaceWithBasis,
)


class NumpyBasedVectorArrayImpl(VectorArrayImpl):

    dim: int

    @abstractmethod
    def create_impl(self, data: NDArray) -> 'NumpyBasedVectorArrayImpl':
        pass

    @abstractmethod
    def __len__(self) -> int:
        pass

    @abstractmethod
    def to_numpy(self, ensure_copy: bool, ind: Indices) -> NDArray:
        pass

    @abstractmethod
    def copy(self, ind: Indices) -> 'NumpyBasedVectorArrayImpl':
        pass

    @abstractmethod
    def append(self, other: 'VectorArrayImpl', remove_from_other: bool, oind: Indices) -> None:
        pass

    @abstractmethod
    def delete(self, ind: Indices) -> None:
        pass

    @abstractmethod
    def set_from_numpy(self, other: NDArray, ind: Indices) -> None:
        pass

    @abstractmethod
    def scal(self, alpha: ArrayLike, ind: Indices) -> None:
        pass

    @abstractmethod
    def axpy_from_numpy(self, alpha: ArrayLike, x: NDArray, ind: Indices) -> None:
        pass

    def is_compatible_array(self, other: 'VectorArrayImpl') -> bool:
        return (isinstance(other, NumpyVectorArrayImpl) and self.dim == other.dim)

    def setitem(self, other: 'VectorArrayImpl', ind: Indices, oind: Indices):
        self.set_from_numpy(other.to_numpy(False, oind), ind)

    def axpy(self, alpha: ArrayLike, x: 'VectorArrayImpl', ind: Indices, xind: Indices) -> None:
        ind = slice(None, self._len) if ind is None else ind
        self.axpy_from_numpy(alpha, x.to_numpy(False, xind), ind)

    def lincomb(self, coefficients: ArrayLike, ind: Indices) -> 'NumpyBasedVectorArrayImpl':
        A = self.to_numpy(False, ind)
        return self.create_impl(coefficients.dot(A))

    def real(self, ind: Indices) -> 'NumpyBasedVectorArrayImpl':
        A = self.to_numpy(False, ind)
        assert not np.isrealobj(A)  # real case handled already by VectorArrayBase
        return self.create_impl(A.real)

    def imag(self, ind: Indices) -> 'NumpyBasedVectorArrayImpl':
        A = self.to_numpy(False, ind)
        return self.create_impl(A.imag)

    def conj(self, ind: Indices) -> 'NumpyBasedVectorArrayImpl':
        A = self.to_numpy(False, ind)
        return self.create_impl(A.conj)

    def dual_pairing(self, other: 'VectorArrayImpl', ind: Indices, oind: Indices, pairwise: bool) -> NDArray:
        A = self.to_numpy(False, ind)
        B = other.to_numpy(False, oind)

        if pairwise:
            # .conj() is a no-op on non-complex data types
            return np.sum(A.conj() * B, axis=1)
        else:
            return A.conj().dot(B.T)


class NumpyVectorArrayImpl(NumpyBasedVectorArrayImpl):
    def __init__(self, array, l=None):
        self._array = array
        self._len = len(array) if l is None else l
        self.scalar_type = array.dtype

    @property
    def dim(self) -> int:
        return self._array.shape[1]

    def create_impl(self, data: NDArray) -> 'NumpyVectorArrayImpl':
        return type(self)(data)

    def __len__(self) -> int:
        return self._len

    def to_numpy(self, ensure_copy, ind) -> NDArray:
        A = self._array[:self._len] if ind is None else self._array[ind]
        if ensure_copy and not A.flags['OWNDATA']:
            return A.copy()
        else:
            return A

    def copy(self, ind: Indices) -> 'NumpyVectorArrayImpl':
        new_array = self._array[:self._len] if ind is None else self._array[ind]
        if not new_array.flags['OWNDATA']:
            new_array = new_array.copy()
        return NumpyVectorArrayImpl(new_array)

    def append(self, other: VectorArrayImpl, remove_from_other: bool, oind: Indices) -> 'NumpyVectorArrayImpl':
        other_array = other.to_numpy(False, oind)
        len_other = len(other_array)
        if len_other == 0:
            return

        if len_other <= self._array.shape[0] - self._len:
            if self._array.dtype != other_array.dtype:
                self._array = self._array.astype(np.promote_types(self._array.dtype, other_array.dtype), copy=False)
            self._array[self._len:self._len + len_other] = other_array
        else:
            self._array = np.append(self._array[:self._len], other_array, axis=0)
        self._len += len_other

        if remove_from_other:
            other.delete(oind)

    def delete(self, ind: Indices) -> None:
        if ind is None:
            self._array = np.empty((0, self._array.shape[1]))
            self._len = 0
            return
        if type(ind) is slice:
            ind = set(range(*ind.indices(self._len)))
        elif not hasattr(ind, '__len__'):
            ind = {ind if 0 <= ind else self._len + ind}
        else:
            l = self._len
            ind = {i if 0 <= i else l+i for i in ind}
        remaining = sorted(set(range(len(self))) - ind)
        self._array = self._array[remaining]
        self._len = len(self._array)
        if not self._array.flags['OWNDATA']:
            self._array = self._array.copy()

    def scal(self, alpha: ArrayLike, ind: Indices) -> None:
        ind = slice(None, self._len) if ind is None else ind
        if type(alpha) is np.ndarray:
            alpha = alpha[:, np.newaxis]

        alpha_type = type(alpha)
        alpha_dtype = alpha.dtype if alpha_type is np.ndarray else alpha_type
        if self._array.dtype != alpha_dtype:
            self._array = self._array.astype(np.promote_types(self._array.dtype, alpha_dtype), copy=False)
        self._array[ind] *= alpha

    def axpy_from_numpy(self, alpha: ArrayLike, x: NDArray, ind: Indices) -> None:
        ind = slice(None, self._len) if ind is None else ind

        alpha_type = type(alpha)
        alpha_dtype = alpha.dtype if alpha_type is np.ndarray else alpha_type
        if self._array.dtype != alpha_dtype or self._array.dtype != x.dtype:
            dtype = np.promote_types(self._array.dtype, alpha_dtype)
            dtype = np.promote_types(dtype, x.dtype)
            self._array = self._array.astype(dtype, copy=False)

        if type(alpha) is np.ndarray:
            alpha = alpha[:, np.newaxis]

        if isinstance(alpha, Number):
            if alpha == 1:
                self._array[ind] += x
                return
            elif alpha == -1:
                self._array[ind] -= x
                return

        self._array[ind] += x * alpha


class NumpyVectorArray(VectorArrayBase):

    impl: NumpyBasedVectorArrayImpl

    def _to_numpy(self, ensure_copy=False) -> NDArray:
        return self.impl.to_numpy(ensure_copy, self.ind)

    def _set_from_numpy(self, other: NDArray) -> None:
        self.impl.set_from_numpy(other, self.ind)

    def _axpy_from_numpy(self, alpha: ArrayLike, x: NDArray) -> None:
        self.impl.axpy_from_numpy(alpha, x, self.ind)

    def __str__(self):
        return str(self._to_numpy())


class NumpyBasedVectorSpace(VectorSpaceWithBasis):

    def __init__(self, dim: int):
        self.dim = int(dim)

    @abstractmethod
    def _make_impl(self, data: NDArray, l: int | None = None) -> NumpyBasedVectorArrayImpl:
        pass

    def empty(self, size_hint=0) -> NumpyVectorArray:
        return NumpyVectorArray(impl=self._make_impl(np.empty((size_hint, self.dim)), l=0))

    def random(self, count: int = 1, distribution: str = 'uniform', seed=None, **kwargs) -> NumpyVectorArray:
        assert count >= 0
        return NumpyVectorArray(
            impl=self._make_impl(_create_random_values((count, self.dim), distribution, seed=seed, **kwargs))
        )

    def from_data(self, data) -> NumpyVectorArray:
        if type(data) is np.ndarray:
            pass
        elif issparse(data):
            data = data.toarray()
        else:
            data = np.array(data, ndmin=2)
        if data.ndim != 2:
            assert data.ndim == 1
            data = np.reshape(data, (1, -1))
        assert data.shape[1] == self.dim
        return NumpyVectorArray(impl=self._make_impl(data))

    def __contains__(self, element: VectorArray) -> bool:
        return isinstance(element, NumpyVectorArray) and element.impl.dim == self.dim

    def __ge__(self, other: 'VectorSpace') -> bool:
        return isinstance(other, NumpyBasedVectorSpace) and self.dim == other.dim

    def from_numpy(self, data, ensure_copy=False) -> NumpyVectorArray:
        if ensure_copy:
            data = data.copy()
        return self.from_data(data)

    def to_numpy(self, U: VectorArray, ensure_copy=False) -> NDArray:
        assert U in self
        return U._to_numpy(ensure_copy)

    def l2_norm(self, U: VectorArray) -> NDArray:
        assert U in self
        return np.linalg.norm(U._to_numpy(), axis=1)

    def amax(self, U: VectorArray) -> (NDArray, NDArray):
        assert U in self
        A = np.abs(U._to_numpy())
        max_ind = np.argmax(A, axis=1)
        max_val = A[np.arange(len(A)), max_ind]
        return max_ind, max_val


class NumpyVectorSpace(NumpyBasedVectorSpace):

    def _make_impl(self, data: NDArray, l: int | None = None) -> NumpyVectorArrayImpl:
        return NumpyVectorArrayImpl(data, l=l)

    @property
    def antidual_space(self) -> 'NumpyVectorSpace':
        return self

    def __eq__(self, other: 'VectorSpace') -> bool:
        return type(other) is type(self) and self.dim == other.dim

    def __hash__(self):
        return hash(self.dim)


def _create_random_values(shape, distribution, seed=None, **kwargs):
    if distribution not in ('uniform', 'normal'):
        raise NotImplementedError

    rng = np.random.default_rng(seed)

    if distribution == 'uniform':
        if not kwargs.keys() <= {'low', 'high'}:
            raise ValueError
        low = kwargs.get('low', 0.)
        high = kwargs.get('high', 1.)
        if high <= low:
            raise ValueError
        return rng.uniform(low, high, shape)
    elif distribution == 'normal':
        if not kwargs.keys() <= {'loc', 'scale'}:
            raise ValueError
        loc = kwargs.get('loc', 0.)
        scale = kwargs.get('scale', 1.)
        return rng.normal(loc, scale, shape)
    else:
        assert False



# This file is part of the NiAS project (https://github.com/nias-project).
# Copyright NiAS developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

from typing import overload

import numpy as np
from numpy.typing import ArrayLike, NDArray

from nias.base.linear_solvers import default_factory
from nias.interfaces import (
    HilbertSpace,
    HSLinearOperator,
    InnerProduct,
    LinearOperator,
    LinearSolver,
    LinearSolverFactory,
    Norm,
    Operator,
    SesquilinearForm,
    VectorArray,
    VectorSpace,
    dual_pairing,
)


@overload
def lincomb(operators: list[Operator], coefficients: ArrayLike) -> Operator:
    ...

@overload
def lincomb(operators: list[LinearOperator], coefficients: ArrayLike) -> LinearOperator:
    ...

@overload
def lincomb(operators: list[HSLinearOperator], coefficients: ArrayLike) -> HSLinearOperator:
    ...


def lincomb(operators, coefficients):
    ops = []
    coeffs = []
    for o, c in zip(operators, coefficients):
        if isinstance(o, LincombOperator):
            ops.extend(o.operators)
            coeffs.extend(o.coefficients * c)
        else:
            ops.append(o)
            coeffs.append(c)

    if all(isinstance(o, HSLinearOperator) for o in ops):
        return HSLinearLincombOperator(ops, coeffs)
    elif all(isinstance(o, LinearOperator) for o in ops):
        return LinearLincombOperator(ops, coeffs)
    else:
        return LincombOperator(ops, coeffs)


class LincombOperator(Operator):
    def __init__(self, operators: list[Operator], coefficients: ArrayLike):
        coefficients = np.array(coefficients)
        assert len(operators) >= 1
        assert coefficients.ndim == 1
        assert len(operators) == len(coefficients)
        assert all(op.source_space == operators[0].source_space for op in operators)
        assert all(op.range_space == operators[0].range_space for op in operators)
        self.operators = list(operators)
        self.coefficients = coefficients
        self.range_space = operators[0].range_space
        self.source_space = operators[0].source_space

    def apply(self, U: VectorArray) -> VectorArray:
        assert U in self.source_space
        if self.coefficients[0] != 0:
            V = self.operators[0].apply(U)
            V.scal(self.coefficients[0])
        else:
            V = self.range_space.zeros(len(U))
        for op, c in zip(self.operators[1:], self.coefficients[1:]):
            V.axpy(c, op.apply(U))
        return V


class LinearLincombOperator(LincombOperator, LinearOperator):

    def __init__(self, operators: list[LinearOperator], coefficients: ArrayLike):
        super().__init__(operators, coefficients)
        assert all(isinstance(o, LinearOperator) for o in operators)

    def apply_transpose(self, V: VectorArray) -> VectorArray:
        assert V in self.range_space.antidual_space
        if self.coefficients[0] != 0:
            U = self.operators[0].apply_transpose(V)
            U.scal(self.coefficients[0])
        else:
            U = self.range_space.zeros(len(V))
        for op, c in zip(self.operators[1:], self.coefficients[1:]):
            U.axpy(c, op.apply_transpose(V))
        return V


class HSLinearLincombOperator(LinearLincombOperator, HSLinearOperator):

    def __init__(self, operators: list[HSLinearOperator], coefficients: ArrayLike):
        super().__init__(operators, coefficients)
        assert all(isinstance(o, HSLinearOperator) for o in operators)

    def apply_adjoint(self, V: VectorArray) -> VectorArray:
        assert V in self.range_space
        if self.coefficients[0] != 0:
            U = self.operators[0].apply_adjoint(V)
            U.scal(self.coefficients[0])
        else:
            U = self.range_space.zeros(len(V))
        for op, c in zip(self.operators[1:], self.coefficients[1:]):
            U.axpy(c, op.apply_adjoint(V))
        return V


@overload
def concat(operators: list[Operator]) -> Operator:
    ...

@overload
def concat(operators: list[LinearOperator]) -> LinearOperator:
    ...

@overload
def concat(operators: list[HSLinearOperator]) -> HSLinearOperator:
    ...


def concat(operators):
    ops = []
    for o in operators:
        if isinstance(o, IdentityOperator):
            pass
        elif isinstance(o, ConcatenationOperator):
            ops.extend(o.operators)
        else:
            ops.append(o)

    if not ops:  # all IdentityOperators
        return operators[0]
    if len(ops) == 1:
        return ops[0]

    if all(isinstance(o, HSLinearOperator) for o in ops):
        return HSConcatenationOperator(ops)
    elif all(isinstance(o, LinearOperator) for o in ops):
        return LinearConcatenationOperator(ops)
    else:
        return ConcatenationOperator(ops)


class ConcatenationOperator(Operator):
    def __init__(self, operators: list[Operator]):
        assert len(operators) >= 1
        assert all(op.source_space >= prev_op.range_space
                   for op, prev_op in zip(operators, operators[1:]))
        self.operators = list(operators)
        self.range_space = operators[0].range_space
        self.source_space = operators[-1].source_space

    def apply(self, U: VectorArray) -> VectorArray:
        assert U in self.source_space
        for op in self.operators[::-1]:
            U = op.apply(U)
        return U


class LinearConcatenationOperator(ConcatenationOperator, LinearOperator):

    def __init__(self, operators: list[LinearOperator]):
        super().__init__(operators)
        assert all(isinstance(o, LinearOperator) for o in operators)

    def apply_transpose(self, V: VectorArray) -> VectorArray:
        assert V in self.range_space.antidual_space
        for op in self.operators:
            V = op.apply_transpose(V)
        return V


class HSConcatenationOperator(LinearConcatenationOperator, HSLinearOperator):

    def __init__(self, operators: list[HSLinearOperator]):
        super().__init__(operators)
        assert all(isinstance(o, HSLinearOperator) for o in operators)

    def apply_adjoint(self, V: VectorArray) -> VectorArray:
        assert V in self.range_space
        for op in self.operators:
            V = op.apply_adjoint(V)
        return V


class OperatorBasedSesquilinearForm(SesquilinearForm):

    def __init__(self, operator: LinearOperator):
        self.operator = operator
        self.source_space = operator.source_space
        self.range_space = operator.range_space.antidual_space

    def apply(self, left: VectorArray, right: VectorArray, pairwise: bool = False) -> NDArray:
        return dual_pairing(left, self.operator.apply(right), pairwise=pairwise)


class OperatorBasedInnerProduct(OperatorBasedSesquilinearForm, InnerProduct):

    pass


class InnerProductBasedNorm(Norm):

    def __init__(self, inner_product: InnerProduct):
        self.inner_product = inner_product

    def __call__(self, U: VectorArray) -> NDArray:
        return np.sqrt(self.inner_product.apply(U, U, pairwise=True))


@overload
def inverse(operator: LinearOperator, context: str = '',
            solver_factory: LinearSolverFactory = default_factory) -> LinearOperator:
    ...

@overload
def inverse(operator: HSLinearOperator, context: str = '',
            solver_factory: LinearSolverFactory = default_factory) -> HSLinearOperator:
    ...


def inverse(operator: Operator, context: str = '', solver_factory: LinearSolverFactory | None = None):
    if solver_factory is None:
        from nias.base.linear_solvers import default_factory
        solver_factory = default_factory

    if isinstance(operator, IdentityOperator):
        return operator

    solver = solver_factory.get_solver(operator, context)

    if isinstance(operator, HSLinearOperator):
        return InverseHSLinearOperator(operator, solver)
    elif isinstance(operator, LinearOperator):
        return InverseLinearOperator(operator, solver)
    else:
        raise NotImplementedError


class InverseLinearOperator(LinearOperator):

    def __init__(self, operator: LinearOperator, solver: LinearSolver):
        self.operator, self.solver = operator, solver
        self.range_space = operator.source_space
        self.source_space = operator.range_space

    def apply(self, U: VectorArray) -> VectorArray:
        assert U in self.source_space
        return self.solver.solve(U)

    def apply_transpose(self, V: VectorArray) -> VectorArray:
        assert V in self.range_space
        return self.solver.solve_transposed(V)


class InverseHSLinearOperator(InverseLinearOperator, HSLinearOperator):
    pass


@overload
def identity(space: VectorSpace) -> Operator:
    ...

@overload
def identity(space: HilbertSpace) -> Operator:
    ...

def identity(space):
    if isinstance(space, HilbertSpace):
        return IdentityOperator(space)
    else:
        return HSIdentityOperator(space)


class IdentityOperator(LinearOperator):

    def __init__(self, space):
        self.space = self.range_space = self.source_space = space

    def apply(self, U: VectorArray) -> VectorArray:
        assert U in self.space
        return U.copy()

    def apply_transpose(self, V: VectorArray) -> VectorArray:
        assert V in self.range_space.antidual_space
        return V.copy()


class HSIdentityOperator(IdentityOperator, HSLinearOperator):

    def apply_adjoint(self, V: VectorArray) -> VectorArray:
        assert V in self.range_space
        return V.copy()



# This file is part of the NiAS project (https://github.com/nias-project).
# Copyright NiAS developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

from functools import reduce

import numpy as np
import scipy.sparse as sps

from nias.bindings.numpy.vectorarrays import NumpyVectorArray, NumpyVectorSpace
from nias.interfaces import LinearOperator, VectorSpaceWithBasis


class NumpyMatrixOperator(LinearOperator):

    sparse: bool

    def __init__(self, matrix,
                 range_space: VectorSpaceWithBasis | None = None,
                 source_space: VectorSpaceWithBasis | None = None):
        assert matrix.ndim <= 2
        if matrix.ndim == 1:
            matrix = np.reshape(matrix, (1, -1))
        if source_space is None:
            source_space = NumpyVectorSpace(matrix.shape[1])
        if range_space is None:
            range_space = NumpyVectorSpace(matrix.shape[0])

        self.matrix = matrix
        self.source_space = source_space
        self.range_space = range_space
        self.sparse = sps.issparse(matrix)

    def apply(self, U: NumpyVectorArray) -> NumpyVectorArray:
        assert U in self.source_space
        return self.range_space.from_numpy((self.matrix @ U._to_numpy().T).T)

    def apply_transpose(self, V: NumpyVectorArray) -> NumpyVectorArray:
        assert V in self.range_space.antidual_space
        return self.source_space.antidual_space.from_numpy((self.matrix.T @ V.to_numpy().T).T)


def assemble_lincomb(operators, coefficients, identity_shift=0.):
    assert all(isinstance(op, NumpyMatrixOperator) for op in operators)

    common_mat_dtype = reduce(np.promote_types,
                              (op.matrix.dtype for op in operators if hasattr(op, 'matrix')))
    common_coef_dtype = reduce(np.promote_types, (type(c) for c in coefficients + [identity_shift]))
    common_dtype = np.promote_types(common_mat_dtype, common_coef_dtype)

    if coefficients[0] == 1:
        matrix = operators[0].matrix.astype(common_dtype)
    else:
        matrix = operators[0].matrix * coefficients[0]
        if matrix.dtype != common_dtype:
            matrix = matrix.astype(common_dtype)

    for op, c in zip(operators[1:], coefficients[1:]):
        if c == 1:
            try:
                matrix += op.matrix
            except NotImplementedError:
                matrix = matrix + op.matrix
        elif c == -1:
            try:
                matrix -= op.matrix
            except NotImplementedError:
                matrix = matrix - op.matrix
        else:
            try:
                matrix += (op.matrix * c)
            except NotImplementedError:
                matrix = matrix + (op.matrix * c)

    if identity_shift != 0:
        if identity_shift.imag == 0:
            identity_shift = identity_shift.real
        if operators[0].sparse:
            try:
                matrix += (sps.eye(matrix.shape[0]) * identity_shift)
            except NotImplementedError:
                matrix = matrix + (sps.eye(matrix.shape[0]) * identity_shift)
        else:
            matrix += (np.eye(matrix.shape[0]) * identity_shift)

    return NumpyMatrixOperator(matrix, source_space=operators[0].source_space, range_space=operators[0].range_space)



# This file is part of the NiAS project (https://github.com/nias-project).
# Copyright NiAS developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

__version__ = '0.0.2'



# This file is part of the NiAS project (https://github.com/nias-project).
# Copyright NiAS developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

from logging import getLogger

import numpy as np
import scipy.linalg as spla

from nias.base.linear_solvers import default_factory
from nias.base.operators import identity, inverse
from nias.interfaces import HSLinearOperator, LinearSolverFactory
from nias.linalg.gram_schmidt import gram_schmidt


def eigs(A: HSLinearOperator, E: HSLinearOperator | None = None,
         k=3, sigma=None, which='LM', b=None, l=None, maxiter=1000, tol=1e-13,
         imag_tol=1e-12, complex_pair_tol=1e-12, complex_evp=False, left_evp=False,
         solver_factory: LinearSolverFactory = default_factory):
    logger = getLogger(__name__)

    assert A.source_space == A.range_space

    if E is None:
        E = identity(A.source_space)
    assert E.source_space == E.range_space
    assert E.source_space == A.source_space

    if b is None:
        b = A.source_space.random()
    assert b in A.source_space

    n = A.source_space.dim

    if l is None:
        l_min = 20
        l = max(2 * k + 1, l_min)
        if n is not None:
            l = min(n - 1, l)

    assert n is None or k < n
    assert l > k

    if sigma is None:
        if left_evp:
            raise NotImplementedError
            # Aop = inverse(E).H @ A.H
        else:
            Aop = inverse(E, context='eigs_E_inverse') @ A
    else:
        if sigma.imag != 0:
            complex_evp = True
        else:
            sigma = sigma.real

        if left_evp:
            raise NotImplementedError
            # Aop = inverse(A - sigma * E).H @ E.H
        else:
            Aop = inverse(A - sigma * E, context='eigs_shift_invert', solver_factory=solver_factory) @ E

    V, H, f = _arnoldi(Aop, k, b, complex_evp)

    k0 = k
    i = 0

    while True:
        i += 1

        V, H, f = _extend_arnoldi(Aop, V, H, f, l - k)

        ew, ev = spla.eig(H)

        # truncate small imaginary parts
        ew.imag[np.abs(ew.imag) / np.abs(ew) < imag_tol] = 0

        if which == 'LM':
            idx = np.argsort(-np.abs(ew))
        elif which == 'SM':
            idx = np.argsort(np.abs(ew))
        elif which == 'LR':
            idx = np.argsort(-ew.real)
        elif which == 'SR':
            idx = np.argsort(ew.real)
        elif which == 'LI':
            idx = np.argsort(-np.abs(ew.imag))
        elif which == 'SI':
            idx = np.argsort(np.abs(ew.imag))

        k = k0
        ews = ew[idx]
        evs = ev[:, idx]

        rres = A.source_space.norm(f)[0] * np.abs(evs[l - 1]) / np.abs(ews)

        # increase k by one in order to keep complex conjugate pairs together
        if not complex_evp and ews[k - 1].imag != 0 and ews[k - 1].imag + ews[k].imag < complex_pair_tol:
            k += 1

        logger.info(f'Maximum of relative Ritz estimates at step {i}: {rres[:k].max():.5e}')

        if np.all(rres[:k] <= tol) or i >= maxiter:
            break

        # increase k in order to prevent stagnation
        k = min(l - 1, k + min(np.count_nonzero(rres[:k] <= tol), (l - k) // 2))

        # sort shifts for QR iteration based on their residual
        shifts = ews[k:l]
        srres = rres[k:l]
        idx = np.argsort(-srres)
        srres = srres[idx]
        shifts = shifts[idx]

        # don't use converged unwanted Ritz values as shifts
        shifts = shifts[srres != 0]
        k += np.count_nonzero(srres == 0)
        if not complex_evp and shifts[0].imag != 0 and shifts[0].imag + ews[1].imag >= complex_pair_tol:
            shifts = shifts[1:]
            k += 1

        H, Qs = _qr_iteration(H, shifts, complex_evp=complex_evp)

        V = V.lincomb(Qs.T)
        f = V[k] * H[k, k - 1] + f * Qs[l - 1, k - 1]
        V = V[:k]
        H = H[:k, :k]

    if sigma is not None:
        ews = 1 / ews + sigma

    return ews[:k0], V.lincomb(evs[:, :k0].T)


def _arnoldi(A, l, b, complex_evp):
    """Compute an Arnoldi factorization."""
    v = b * (1 / A.source_space.norm(b)[0])

    H = np.zeros((l, l), dtype=np.complex_ if complex_evp else np.float_)
    V = A.source_space.empty(size_hint=l)

    V.append(v)

    for i in range(l):
        v = A.apply(v)
        V.append(v)

        _, R = gram_schmidt(V, A.source_space.inner_product,
                            return_R=True, atol=0, rtol=0, offset=len(V) - 1, copy=False)
        H[:i + 2, i] = R[:l, i + 1]
        v = V[-1]

    return V[:l], H, v * R[l, l]


def _extend_arnoldi(A, V, H, f, p):
    """Extend an existing Arnoldi factorization."""
    k = len(V)

    res = A.source_space.norm(f)[0]
    # the explicit "constant" mode is needed for numpy 1.16
    # mode only gained a default value with numpy 1.17
    H = np.pad(H, ((0, p), (0, p)), mode='constant')
    H[k, k - 1] = res
    v = f * (1 / res)
    V = V.copy()
    V.append(v)

    for i in range(k, k + p):
        v = A.apply(v)
        V.append(v)
        _, R = gram_schmidt(V, A.source_space.inner_product,
                            return_R=True, atol=0, rtol=0, offset=len(V) - 1, copy=False)
        H[:i + 2, i] = R[:k + p, i + 1]

        v = V[-1]

    return V[:k + p], H, v * R[k + p, k + p]


def _qr_iteration(H, shifts, complex_evp=False):
    """Perform the QR iteration."""
    Qs = np.eye(len(H))

    i = 0
    while i < len(shifts) - 1:
        s = shifts[i]
        if not complex_evp and shifts[i].imag != 0:
            Q, _ = spla.qr(H @ H - 2 * s.real * H + np.abs(s)**2 * np.eye(len(H)))
            i += 2
        else:
            Q, _ = spla.qr(H - s * np.eye(len(H)))
            i += 1
        Qs = Qs @ Q
        H = Q.T @ H @ Q

    return H, Qs



# This file is part of the NiAS project (https://github.com/nias-project).
# Copyright NiAS developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

from logging import getLogger

import numpy as np

from nias.exceptions import AccuracyError
from nias.interfaces import InnerProduct, VectorArray


def gram_schmidt(A: VectorArray, inner_product: InnerProduct,
                 return_R=False, atol=1e-13, rtol=1e-13, offset=0,
                 reiterate=True, reiteration_threshold=9e-1, check=True, check_tol=1e-3,
                 copy=True):

    logger = getLogger(__name__)

    if copy:
        A = A.copy()

    norm = inner_product.induced_norm()

    # main loop
    R = np.eye(len(A), dtype=A.scalar_type)
    remove = []  # indices of to be removed vectors
    for i in range(offset, len(A)):
        # first calculate norm
        initial_norm = norm(A[i])[0]

        if initial_norm <= atol:
            logger.info(f'Removing vector {i} of norm {initial_norm}')
            remove.append(i)
            continue

        if i == 0:
            A[0].scal(1 / initial_norm)
            R[i, i] = initial_norm
        else:
            current_norm = initial_norm
            # If reiterate is True, reiterate as long as the norm of the vector changes
            # strongly during orthogonalization (due to Andreas Buhr).
            while True:
                # orthogonalize to all vectors left
                for j in range(i):
                    if j in remove:
                        continue
                    p = inner_product(A[j], A[i])[0, 0]
                    A[i].axpy(-p, A[j])
                    R[j, i] += p

                # calculate new norm
                old_norm, current_norm = current_norm, norm(A[i])[0]

                # remove vector if it got too small
                if current_norm <= rtol * initial_norm:
                    logger.info(f'Removing linearly dependent vector {i}')
                    remove.append(i)
                    break

                # check if reorthogonalization should be done
                if reiterate and current_norm < reiteration_threshold * old_norm:
                    logger.info(f'Orthonormalizing vector {i} again')
                else:
                    A[i].scal(1 / current_norm)
                    R[i, i] = current_norm
                    break

    if remove:
        del A[remove]
        R = np.delete(R, remove, axis=0)

    if check:
        error_matrix = inner_product(A[offset:len(A)], A)
        error_matrix[:len(A) - offset, offset:len(A)] -= np.eye(len(A) - offset)
        if error_matrix.size > 0:
            err = np.max(np.abs(error_matrix))
            if err >= check_tol:
                raise AccuracyError(f'result not orthogonal (max err={err})')

    if return_R:
        return A, R
    else:
        return A



# This file is part of the NiAS project (https://github.com/nias-project).
# Copyright NiAS developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

import numpy as np
from scipy.linalg import lu_factor, lu_solve
from scipy.linalg.lapack import get_lapack_funcs
from scipy.sparse.linalg import splu, spsolve

from nias.base.linear_solvers import default_factory
from nias.base.operators import IdentityOperator, LincombOperator
from nias.bindings.numpy.operators import NumpyMatrixOperator, assemble_lincomb
from nias.exceptions import InversionError
from nias.interfaces import LinearOperator, LinearSolver, VectorArray


class DirectSolver(LinearSolver):

    def __init__(self, lhs: LinearOperator, keep_factorization: bool,
                 permc_spec: str, check_finite: bool, check_cond: bool):
        self.keep_factorization, self.permc_spec, self.check_finite, self.check_cond = \
            keep_factorization, permc_spec, check_finite, check_cond
        self.factorization = None
        self.set_lhs(lhs)

    def set_lhs(self, lhs: LinearOperator) -> None:
        if isinstance(lhs, LincombOperator):
            # TODO: move this somewhere else
            assert all(isinstance(o, (IdentityOperator, NumpyMatrixOperator)) for o in lhs.operators)
            identity_shift = 0
            ops, coeffs = [], []
            for o, c in zip(lhs.operators, lhs.coefficients):
                if isinstance(o, IdentityOperator):
                    identity_shift += c
                else:
                    ops.append(o)
                    coeffs.append(c)
            if not ops:
                raise NotImplementedError
            lhs = assemble_lincomb(ops, coeffs, identity_shift=identity_shift)
        assert isinstance(lhs, NumpyMatrixOperator)
        self.lhs = lhs

    def solve(self, rhs: VectorArray) -> VectorArray:
        assert rhs in self.lhs.range_space
        matrix = self.lhs.matrix
        V = self.lhs.range_space.to_numpy(rhs)
        promoted_type = np.promote_types(matrix.dtype, V.dtype)

        if not self.lhs.sparse:
            if self.factorization is None:
                try:
                    self.factorization = lu_factor(matrix, check_finite=self.check_finite)
                except np.linalg.LinAlgError as e:
                    raise InversionError(f'{type(e)!s}: {e!s}') from e
                if self.check_cond:
                    gecon = get_lapack_funcs('gecon', self.factorization)
                    rcond, _ = gecon(self.factorization[0], np.linalg.norm(matrix, ord=1), norm='1')
                    if rcond < np.finfo(np.float64).eps:
                        self.logger.warning(f'Ill-conditioned matrix (rcond={rcond:.6g}) in apply_inverse: '
                                            'result may not be accurate.')
            R = lu_solve(self.factorization, V.T, check_finite=self.check_finite).T
            return self.lhs.source_space.from_numpy(R)

        try:
            # maybe remove unusable factorization:
            if self.factorization is not None:
                fdtype = self.factorizationdtype
                if not np.can_cast(V.dtype, fdtype, casting='safe'):
                    self.factorization = None

            if self.factorization is not None:
                # we may use a complex factorization of a real matrix to
                # apply it to a real vector. In that case, we downcast
                # the result here, removing the imaginary part,
                # which should be zero.
                R = self.factorization.solve(V.T).T.astype(promoted_type, copy=False)
            elif self.keep_factorization:
                # the matrix is always converted to the promoted type.
                # if matrix.dtype == promoted_type, this is a no_op
                self.factorization = splu(matrix_astype_nocopy(matrix.tocsc(), promoted_type),
                                          permc_spec=self.permc_spec)
                self.factorizationdtype = promoted_type
                R = self.factorization.solve(V.T).T
            else:
                # the matrix is always converted to the promoted type.
                # if matrix.dtype == promoted_type, this is a no_op
                R = spsolve(matrix_astype_nocopy(matrix, promoted_type), V.T, permc_spec=self.permc_spec).T
        except RuntimeError as e:
            raise InversionError(e) from e

        return self.lhs.source_space.from_numpy(R)

    def solve_transposed(self, rhs: VectorArray) -> VectorArray:
        pass



default_factory.register_solver(
    (NumpyMatrixOperator, IdentityOperator), 'scipy-spsolve', '10',
    DirectSolver,
    {'keep_factorization': True, 'permc_spec': 'COLAMD', 'check_finite': True, 'check_cond': True}
)


# unfortunately, this is necessary, as scipy does not
# forward the copy=False argument in its csc_matrix.astype function
def matrix_astype_nocopy(matrix, dtype):
    if matrix.dtype == dtype:
        return matrix
    else:
        return matrix.astype(dtype)



# This file is part of the NiAS project (https://github.com/nias-project).
# Copyright NiAS developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)


class AccuracyError(Exception):
    """Is raised if the result of a computation is inaccurate."""


class InversionError(Exception):
    pass



# This file is part of the NiAS project (https://github.com/nias-project).
# Copyright NiAS developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

from typing import Any

from nias.exceptions import InversionError
from nias.interfaces import LinearOperator, LinearSolver, LinearSolverFactory


class DefaultLinearSolverFactory(LinearSolverFactory):

    solvers: list[tuple[int, tuple[type], str, type[LinearSolver]]]

    def __init__(self):
        self.solvers = []
        self.defaults = {}

    def register_solver(self, supported_types: tuple[type], name: str, priority: int,
                        solver: type[LinearSolver], defaults: dict[str, Any]):
        if name in self.defaults:
            raise ValueError(f'Solver with name {name} has already been registered.')
        self.solvers.append((priority, supported_types, name, solver))
        self.defaults[name] = defaults

    def get_solver(self, lhs: LinearOperator, context: str = '') -> LinearSolver:
        from nias.base.operators import LincombOperator

        if isinstance(lhs, LincombOperator):
            base_types = tuple(type(o) for o in lhs.operators)
        else:
            base_types = (type(lhs),)

        for priority, supported_types, name, solver in self.solvers:
            if all(issubclass(bt, supported_types) for bt in base_types):
                return solver(lhs, **self.defaults[name])

        raise InversionError(f'No solver known for base types {base_types}')

    def set_defaults(self, name: str, **values) -> None:
        self.defaults[name].update(values)

    def set_priority(self, name: str, priority: int) -> None:
        for i in range(len(self.solvers)):
            if self.solvers[i][2] == name:
                self.solvers[i][0] = priority
        self._sort_solvers()

    def _sort_solvers(self) -> None:
        self.solvers.sort(key=lambda x: x[0], reverse=True)

    def __str__(self):
        return ('Registered solvers\n------------------\n'
                + '\n'.join(f'{priority} [{", ".join(t.__name__ for t in types)}] {name} {self.defaults[name]}'
                            for priority, types, name, _ in self.solvers))


default_factory = DefaultLinearSolverFactory()



# This file is part of the NiAS project (https://github.com/nias-project).
# Copyright NiAS developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

from abc import ABC, abstractmethod, abstractproperty
from numbers import Number
from typing import Self, TypeAlias, Union

import numpy as np
from numpy.typing import ArrayLike, NDArray

Scalar: TypeAlias = int | float | complex | np.number
Indices: TypeAlias = None | int | slice | list[int] | NDArray


class VectorArray(ABC):
    """Abstract VectorArray interface.

    The scalar_type of a VectorArray never changes with one exception:
    if a real array is multiplied (in-place or not) with a complex scalar,
    its scalar_type changes to the corresponding complex type.
    Not all VectorArray implementations need to support complex numbers,
    however.

    append/axpy/dual_pairing/__add__/__iadd_/__setitem__ require
    both arrays to be compatible. In particular this means that both
    arrays need to be contained in a common (algebraic) vector space.
    Further, the scalar_types of both arrays need to agree
    (up to complexification).

    All complex VectorArrays are assumed to be equiped with a canonical
    conjugation operation, which yields the real and imaginary part of
    a vector in the array.
    """

    base: Union['VectorArray', None]
    ind: Indices
    is_view: bool
    scalar_type: type

    @abstractmethod
    def __len__(self) -> int:
        pass

    @abstractmethod
    def is_compatible_array(self, other: 'VectorArray') -> bool:
        pass

    @abstractmethod
    def copy(self) -> Self:
        pass

    @abstractmethod
    def append(self, other: 'VectorArray', remove_from_other: bool = False) -> None:
        pass

    @abstractmethod
    def __getitem__(self, ind: Indices) -> 'VectorArray':
        pass

    @abstractmethod
    def __setitem__(self, ind: Indices, other: 'VectorArray') -> None:
        pass

    @abstractmethod
    def __delitem__(self, ind: Indices):
        pass

    @abstractmethod
    def scal(self, alpha: ArrayLike) -> None:
        pass

    @abstractmethod
    def axpy(self, alpha: ArrayLike, x: 'VectorArray') -> None:
        pass

    @abstractmethod
    def lincomb(self, coefficients: ArrayLike) -> 'VectorArray':
        pass

    @property
    def is_complex(self) -> bool:
        return isinstance(self.scalar_type, np.complexfloating)

    @abstractproperty
    def real(self) -> 'VectorArray':
        pass

    @abstractproperty
    def imag(self) -> 'VectorArray':
        pass

    @abstractmethod
    def conj(self) -> 'VectorArray':
        pass

    def _dual_pairing(self, other: 'VectorArray') -> NDArray:
        return NotImplemented

    def __add__(self, other: 'VectorArray'):
        result = self.copy()
        result.axpy(1., other)
        return result

    def __iadd__(self, other):
        self.axpy(1, other)
        return self

    __radd__ = __add__

    def __sub__(self, other):
        result = self.copy()
        result.axpy(-1., other)
        return result

    def __isub__(self, other):
        self.axpy(-1, other)
        return self

    def __mul__(self, other):
        result = self.copy()
        result.scal(other)
        return result

    __rmul__ = __mul__

    def __imul__(self, other):
        self.scal(other)
        return self

    def __neg__(self):
        result = self.copy()
        result.scal(-1.)
        return result

    # override NumPy binary operations and ufuncs
    __array_priority__ = 100.0
    __array_ufunc__ = None

    def check_ind(self, ind: Indices) -> bool:
        """Check if index is admissible.

        Check if `ind` is an admissible list of indices in the sense
        of the class documentation.
        """
        l = len(self)
        return (type(ind) is slice
                or isinstance(ind, Number) and -l <= ind < l
                or isinstance(ind, (list, np.ndarray)) and all(-l <= i < l for i in ind))

    def check_ind_unique(self, ind: Indices) -> bool:
        """Check if index is admissible and unique.

        Check if `ind` is an admissible list of non-repeated indices in
        the sense of the class documentation.
        """
        l = len(self)
        return (type(ind) is slice
                or isinstance(ind, Number) and -l <= ind < l
                or isinstance(ind, (list, np.ndarray))
                and len({i if i >= 0 else l+i for i in ind if -l <= i < l}) == len(ind))


def dual_pairing(left: VectorArray, right: VectorArray, pairwise: bool = False) -> NDArray:
    result = left._dual_pairing(right, pairwise)
    if result is NotImplemented:
        result = right._dual_pairing(left, pairwise)
    if result is NotImplemented:
        raise NotImplementedError('No dual pairing possible.')
    return result


class VectorSpace(ABC):
    dim: int | None

    @abstractmethod
    def empty(self, size_hint=0) -> VectorArray:
        pass

    @abstractmethod
    def zeros(self, count: int = 1) -> VectorArray:
        pass

    @abstractmethod
    def random(self, count: int = 1) -> VectorArray:
        pass

    @abstractmethod
    def from_data(self, data) -> VectorArray:
        pass

    @abstractproperty
    def antidual_space(self) -> 'VectorSpace':
        pass

    @abstractmethod
    def __contains__(self, element: 'VectorArray') -> bool:
        pass

    @abstractmethod
    def __eq__(self, other: 'VectorSpace') -> bool:
        pass

    @abstractmethod
    def __ge__(self, other: 'VectorSpace') -> bool:
        """`True` if `other` embeds into `self`."""

    def __le__(self, other: 'VectorSpace') -> bool:
        assert isinstance(other, VectorSpace)
        return other >= self


class VectorSpaceWithBasis(VectorSpace):
    """Vector space with a chosen basis.

    `antidual_space` is assumed to be equipped with antidual basis.
    """

    dim: int

    @abstractmethod
    def from_numpy(self, data, ensure_copy=False) -> VectorArray:
        pass

    @abstractmethod
    def to_numpy(self, U: VectorArray, ensure_copy=False) -> NDArray:
        pass

    @abstractmethod
    def l2_norm(self, U: VectorArray) -> NDArray:
        pass

    @abstractmethod
    def amax(self, U: VectorArray) -> (NDArray, NDArray):
        pass

    def full(self, value, count=1):
        return self.from_numpy(np.full((self.dim, count), value))

    def zeros(self, count: int = 1) -> VectorArray:
        return self.full(0., count)

    def ones(self, count: int = 1) -> VectorArray:
        return self.full(1., count)


class Norm(ABC):
    space: VectorSpace

    @abstractmethod
    def __call__(self, U: VectorArray) -> NDArray:
        pass


class NormedSpace(VectorSpace):
    norm: Norm


class NormedSpaceWithBasis(NormedSpace, VectorSpaceWithBasis):
    pass


class SesquilinearForm(ABC):
    left_space: VectorSpace
    right_space: VectorSpace

    @abstractmethod
    def apply(self, left: VectorArray, right: VectorArray, pairwise: bool = False) -> NDArray:
        pass

    def __call__(self, left: VectorArray, right: VectorArray) -> NDArray:
        return self.apply(left, right)


class InnerProduct(SesquilinearForm):

    def induced_norm(self) -> Norm:
        from nias.base.operators import InnerProductBasedNorm
        return InnerProductBasedNorm(self)


class HilbertSpace(NormedSpace):
    inner_product: InnerProduct

    @abstractmethod
    def riesz(self, U: VectorArray) -> VectorArray:
        pass


class HilbertSpaceWithBasis(HilbertSpace, VectorSpaceWithBasis):
    pass


class EuclideanSpace(HilbertSpaceWithBasis):
    """`inner_product` is compatible with `to_numpy`.

    For complex spaces, `riesz` is not the identity but complex conjugation!
    """

    @property
    def antidual_space(self) -> 'EuclideanSpace':
        return self


class Operator(ABC):
    source_space: VectorSpace
    range_space: VectorSpace

    @abstractmethod
    def apply(self, U: VectorArray) -> VectorArray:
        pass

    def __add__(self, other: 'Operator') -> 'Operator':
        from nias.base.operators import lincomb
        return lincomb([self, other], [1., 1.])

    def __sub__(self, other: 'Operator') -> 'Operator':
        from nias.base.operators import lincomb
        return lincomb([self, other], [1., -1.])

    def __mul__(self, other: Scalar) -> 'Operator':
        from nias.base.operators import lincomb
        return lincomb([self], [other])

    __rmul__ = __mul__

    def __matmul__(self, other: 'Operator') -> 'Operator':
        from nias.base.operators import concat
        return concat([self, other])

    def assemble(self) -> 'Operator':
        return self


class LinearOperator(Operator):

    @abstractmethod
    def apply_transpose(self, V: VectorArray) -> VectorArray:
        pass

    def as_sesquilinear_form(self) -> SesquilinearForm:
        from nias.base.operators import OperatorBasedSesquilinearForm
        return OperatorBasedSesquilinearForm(self)

    def as_inner_product(self) -> InnerProduct:
        from nias.base.operators import OperatorBasedInnerProduct
        return OperatorBasedInnerProduct(self)


class HSLinearOperator(LinearOperator):
    source_space: HilbertSpace
    range_space: HilbertSpace

    def apply_adjoint(self, V: VectorArray) -> VectorArray:
        return self.source_space.dual_space.riesz(self.apply_transpose(self.range.riesz(V)))


class LinearSolver(ABC):

    @abstractmethod
    def __init__(self, lhs: LinearOperator, options: dict):
        pass

    @abstractmethod
    def set_lhs(self, lhs: LinearOperator) -> None:
        pass

    @abstractmethod
    def solve(self, rhs: VectorArray) -> VectorArray:
        pass

    @abstractmethod
    def solve_transposed(self, rhs: VectorArray) -> VectorArray:
        pass


class LinearSolverFactory(ABC):

    @abstractmethod
    def get_solver(self, lhs: LinearOperator, context: str = '') -> LinearSolver:
        pass



# This file is part of the NiAS project (https://github.com/nias-project).
# Copyright NiAS developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

from abc import abstractmethod
from numbers import Integral
from typing import Self

import numpy as np
from numpy.typing import NDArray

from nias.base.linear_solvers import default_factory
from nias.interfaces import (
    ArrayLike,
    HilbertSpace,
    HilbertSpaceWithBasis,
    Indices,
    LinearOperator,
    LinearSolverFactory,
    Scalar,
    VectorArray,
    VectorSpace,
    VectorSpaceWithBasis,
)


class VectorArrayBase(VectorArray):

    base = None
    is_view: bool = False
    _impl = None

    def __init__(self, *, impl=None, base=None, ind=None, _len=None):
        assert ind is None or base is not None
        self.scalar_type = impl.scalar_type
        self.real_scalar_type = impl.real_scalar_type
        self.ind = ind
        if base is None:
            self._impl = impl
            self._refcount = [1]
            self._len = len(impl)
        else:
            self.is_view = True
            self.base = base
            self._len = _len

    @property
    def impl(self):
        return self.base._impl if self.is_view else self._impl

    def __len__(self) -> int:
        return self._len

    def is_compatible_array(self, other: VectorArray) -> bool:
        if self.real_scalar_type != other.real_scalar_type:
            return False
        return self.impl.is_compatible_array(other.impl)

    def copy(self) -> Self:
        if self.is_view:
            return type(self)(impl=self.impl.copy(self.ind))
        else:
            C = type(self)(impl=self.impl)
            C._refcount = self._refcount
            self._refcount[0] += 1
            return C

    def append(self, other: VectorArray, remove_from_other: bool = False) -> None:
        assert isinstance(other, VectorArrayBase)
        assert self.is_compatible_array(other)
        if self.is_view:
            raise ValueError('Cannot append to VectorArray view')
        if remove_from_other and self.impl is other.impl:
            raise ValueError('Cannot append VectorArray to itself with remove_from_other=True')
        self._copy_impl_if_multiple_refs()
        if remove_from_other:
            other._copy_impl_if_multiple_refs()
        self.impl.append(other.impl, remove_from_other, other.ind)
        self._len += other._len
        if remove_from_other:
            if other.is_view:
                other.base._len = len(other.base._impl)  # update _len from impl
            else:
                other._len = 0

    def __getitem__(self, ind: ArrayLike) -> VectorArray:
        l = self._len

        # normalize ind s.t. the length of the view does not change when
        # the array is appended to
        if isinstance(ind, Integral):
            if 0 <= ind < l:
                ind = slice(ind, ind+1)
            elif ind >= l or ind < -l:
                raise IndexError('VectorArray index out of range')
            else:
                ind = l+ind
                ind = slice(ind, ind+1)
            view_len = 1
        elif type(ind) is slice:
            start, stop, step = ind.indices(l)
            if start == stop:
                ind = slice(0, 0, 1)
                view_len = 0
            else:
                assert start >= 0
                assert stop >= 0 or (step < 0 and stop >= -1)
                ind = slice(start, None if stop == -1 else stop, step)
                view_len = len(range(start, stop, step))
        else:
            assert isinstance(ind, (list, np.ndarray))
            assert all(-l <= i < l for i in ind)
            ind = [i if 0 <= i else l+i for i in ind]
            view_len = len(ind)

        if self.is_view:
            ind = self._sub_index(self.base._len, self.ind, ind)
            return type(self)(impl=self.impl, base=self.base, ind=ind, _len=view_len)
        else:
            return type(self)(impl=self.impl, base=self, ind=ind, _len=view_len)

    def __setitem__(self, indices: ArrayLike, other: VectorArray) -> None:
        assert not self.is_view or self.base.check_ind_unique(self.ind)
        assert isinstance(other, VectorArrayBase)
        assert self.is_compatible_array(other)
        assert len(self) == len(other) or len(other) == 1
        self._copy_impl_if_multiple_refs()
        self.impl.setitem(other.impl, self.ind, other.ind)

    def __delitem__(self, ind: ArrayLike):
        if self.is_view:
            raise ValueError('Cannot delete items from VectorArray view')
        assert self.check_ind(ind)
        self._copy_impl_if_multiple_refs()
        self.impl.delete(ind)
        self._len = len(self._impl)

    def scal(self, alpha: ArrayLike) -> None:
        assert not self.is_view or self.base.check_ind_unique(self.ind)
        assert isinstance(alpha, Scalar) \
            or isinstance(alpha, np.ndarray) and alpha.shape == (len(self),)
        self._copy_impl_if_multiple_refs()
        self.impl.scal(alpha, self.ind)

    def axpy(self, alpha: ArrayLike, x: VectorArray) -> None:
        assert not self.is_view or self.base.check_ind_unique(self.ind)
        assert isinstance(alpha, Scalar) \
            or isinstance(alpha, np.ndarray) and alpha.shape == (len(self),)
        assert isinstance(x, VectorArrayBase)
        assert self.is_compatible_array(x)
        assert len(self) == len(x) or len(x) == 1
        self._copy_impl_if_multiple_refs()
        self.impl.axpy(alpha, x.impl, self.ind, x.ind)

    def lincomb(self, coefficients: ArrayLike) -> VectorArray:
        assert 1 <= coefficients.ndim <= 2
        if coefficients.ndim == 1:
            coefficients = coefficients[np.newaxis, ...]
        assert coefficients.shape[-1] == len(self)
        return type(self)(impl=self.impl.lincomb(coefficients, self.ind))

    @property
    def real(self) -> VectorArray:
        if not self.is_complex:
            return self.copy()
        impl = self.impl.real(self.ind)
        return type(self)(impl=impl)

    @property
    def imag(self) -> VectorArray:
        return type(self)(impl=self.impl.imag(self.ind))

    def conj(self) -> VectorArray:
        if not self.is_complex:
            return self.copy()
        impl = self.impl.conj(self.ind)
        return type(self)(impl=impl)

    def _dual_pairing(self, other: VectorArray, pairwise: bool) -> NDArray:
        assert isinstance(other, VectorArrayBase)
        assert self.is_compatible_array(other)
        assert not pairwise or len(self) == len(other)
        return self.impl.dual_pairing(other.impl, self.ind, other.ind, pairwise)

    def _copy_impl_if_multiple_refs(self):
        array = self.base if self.is_view else self
        if array._refcount[0] == 1:
            return
        array._impl = array._impl.copy(None)  # copy the array implementation
        array._refcount[0] -= 1               # decrease refcount for original array
        array._refcount = [1]                 # create new reference counter

    @staticmethod
    def _sub_index(l, ind, ind_ind):
        if type(ind) is slice:
            ind = range(*ind.indices(l))
            if type(ind_ind) is slice:
                result = ind[ind_ind]
                return slice(result.start, result.stop, result.step)
            elif hasattr(ind_ind, '__len__'):
                return [ind[i] for i in ind_ind]
            else:
                return [ind[ind_ind]]
        else:
            if not hasattr(ind, '__len__'):
                ind = [ind]
            if type(ind_ind) is slice:
                return ind[ind_ind]
            elif hasattr(ind_ind, '__len__'):
                return [ind[i] for i in ind_ind]
            else:
                return [ind[ind_ind]]

    def __del__(self):
        if not self.is_view:
            self._refcount[0] -= 1


class VectorArrayImpl:
    scalar_type: type

    @property
    def real_scalar_type(self) -> type:
        return np.empty(0, dtype=self.scalar_type).real.dtype

    @abstractmethod
    def __len__(self) -> int:
        pass

    @abstractmethod
    def is_compatible_array(self, other: 'VectorArrayImpl') -> bool:
        pass

    @abstractmethod
    def copy(self, ind: Indices) -> 'VectorArrayImpl':
        pass

    @abstractmethod
    def append(self, other: 'VectorArrayImpl', remove_from_other: bool, oind: Indices) -> None:
        pass

    @abstractmethod
    def delete(self, ind: Indices) -> None:
        pass

    @abstractmethod
    def setitem(self, other: 'VectorArrayImpl', ind: Indices, oind: Indices) -> None:
        pass

    @abstractmethod
    def scal(self, alpha: ArrayLike, ind: Indices) -> None:
        pass

    @abstractmethod
    def axpy(self, alpha: ArrayLike, x: 'VectorArrayImpl', ind: Indices, xind: Indices) -> None:
        pass

    @abstractmethod
    def lincomb(self, coefficients: ArrayLike, ind: Indices) -> 'VectorArrayImpl':
        pass

    @abstractmethod
    def real(self, ind: Indices) -> 'VectorArrayImpl':
        pass

    @abstractmethod
    def imag(self, ind: Indices) -> 'VectorArrayImpl':
        pass

    @abstractmethod
    def conj(self, ind: Indices) -> 'VectorArrayImpl':
        pass

    @abstractmethod
    def dual_pairing(self, other: 'VectorArrayImpl', ind: Indices, oind: Indices, pairwise: bool) -> NDArray:
        pass


class HilbertSpaceFromProductOperator(HilbertSpace):

    def __init__(self, algebraic_space: VectorSpace, product_operator: LinearOperator,
                 dual_product_operator: LinearOperator | None = None,
                 solver_factory: LinearSolverFactory = default_factory):
        assert product_operator.source_space >= algebraic_space
        assert product_operator.range_space <= algebraic_space.antidual_space
        assert algebraic_space.antidual_space >= product_operator.range_space
        if dual_product_operator is None:
            from nias.base.operators import inverse
            dual_product_operator = inverse(product_operator, context='dual_product', solver_factory=solver_factory)
        assert dual_product_operator.source_space >= algebraic_space.antidual_space
        assert dual_product_operator.range_space <= algebraic_space
        self.algebraic_space, self.product_operator, self.dual_product_operator \
            = algebraic_space, product_operator, dual_product_operator
        from nias.base.operators import OperatorBasedInnerProduct
        self.inner_product = OperatorBasedInnerProduct(product_operator)
        self.norm = self.inner_product.induced_norm()
        self.dim = algebraic_space.dim

    def empty(self, size_hint=0) -> VectorArray:
        return self.algebraic_space.empty(size_hint)

    def zeros(self, count: int = 1) -> VectorArray:
        return self.algebraic_space.zeros(count)

    def random(self, count: int = 1) -> VectorArray:
        return self.algebraic_space.random(count)

    def from_data(self, data) -> VectorArray:
        return self.algebraic_space.from_data(data)

    @property
    def antidual_space(self):
        return type(self)(self.algebraic_space, self.dual_product_operator, self.product_operator)

    def __contains__(self, element: 'VectorArray') -> bool:
        return element in self.algebraic_space

    def __eq__(self, other: 'VectorSpace') -> bool:
        return isinstance(other, HilbertSpaceFromProductOperator) \
            and other.algebraic_space == self.algebraic_space \
            and other.product_operator == self.product_operator

    def __ge__(self, other: 'VectorSpace') -> bool:
        assert isinstance(other, VectorSpace)
        try:
            other = other.algebraic_space
        except AttributeError:
            pass
        return self.algebraic_space >= other

    def riesz(self, U: VectorArray) -> VectorArray:
        assert U in self.algebraic_space
        return self.product_operator.apply(U)


class HilbertSpaceWithBasisFromProductOperator(HilbertSpaceFromProductOperator, HilbertSpaceWithBasis):

    def __init__(self, algebraic_space: VectorSpaceWithBasis, product_operator: LinearOperator,
                 dual_product_operator: LinearOperator | None = None,
                 solver_factory: LinearSolverFactory = default_factory):
        super().__init__(algebraic_space, product_operator, dual_product_operator, solver_factory)

    def from_numpy(self, data, ensure_copy=False) -> VectorArray:
        return self.algebraic_space.from_numpy(data, ensure_copy=ensure_copy)

    def to_numpy(self, U: VectorArray, ensure_copy=False) -> NDArray:
        return self.algebraic_space.to_numpy(U, ensure_copy=ensure_copy)

    def l2_norm(self, U: VectorArray) -> NDArray:
        return self.algebraic_space.l2_norm(U)

    def amax(self, U: VectorArray) -> (NDArray, NDArray):
        return self.algebraic_space.amax(U)



# This file is part of the NiAS project (https://github.com/nias-project).
# Copyright NiAS developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

from logging import getLogger

import numpy as np
import scipy.linalg as spla

from nias.interfaces import HilbertSpace, VectorArray
from nias.linalg.gram_schmidt import gram_schmidt


def method_of_snapshots(A: VectorArray, range_space: HilbertSpace,
                        modes=None, rtol=1e-7, atol=0., l2_err=0.):
    assert A in range_space

    if range_space.dim == 0 or len(A) == 0:
        return range_space.empty(), np.array([]), np.zeros((0, len(A)))

    logger = getLogger(__name__ + '.method_of_snapshots')

    logger.info(f'Computing Gramian ({len(A)} vectors) ...')
    B = range_space.inner_product(A, A)

    logger.info('Computing eigenvalue decomposition ...')
    eigvals = (None
               if modes is None or l2_err > 0.
               else (max(len(B) - modes, 0), len(B) - 1))

    evals, V = spla.eigh(B, overwrite_a=True, subset_by_index=eigvals)
    evals = evals[::-1]
    V = V.T[::-1, :]

    tol = max(rtol ** 2 * evals[0], atol ** 2)
    above_tol = np.where(evals >= tol)[0]
    if len(above_tol) == 0:
        return range_space.empty(), np.array([]), np.zeros((0, len(A)))
    last_above_tol = above_tol[-1]

    errs = np.concatenate((np.cumsum(evals[::-1])[::-1], [0.]))
    below_err = np.where(errs <= l2_err**2)[0]
    first_below_err = below_err[0]

    selected_modes = min(first_below_err, last_above_tol + 1)
    if modes is not None:
        selected_modes = min(selected_modes, modes)

    if selected_modes > range_space.dim:
        logger.warning('Number of computed singular vectors larger than array dimension! Truncating ...')
        selected_modes = range_space.dim

    s = np.sqrt(evals[:selected_modes])
    V = V[:selected_modes]
    Vh = V.conj()

    logger.info(f'Computing left-singular vectors ({len(V)} vectors) ...')
    U = A.lincomb(V / s[:, np.newaxis])

    return U, s, Vh


def qr_svd(A: VectorArray, range_space: HilbertSpace,
           modes=None, rtol=4e-8, atol=0., l2_err=0.):
    # TODO: add a way to pass gram_schmidt options
    assert A in range_space

    if range_space.dim == 0 or len(A) == 0:
        return range_space.empty(), np.array([]), np.zeros((0, len(A)))

    logger = getLogger(__name__ + '.qr_svd')

    logger.info('Computing QR decomposition ...')
    Q, R = gram_schmidt(A, range_space.inner_product, return_R=True, check=False)

    logger.info('Computing SVD of R ...')
    U2, s, Vh = spla.svd(R, lapack_driver='gesvd')

    logger.info('Choosing the number of modes ...')
    tol = max(rtol * s[0], atol)
    above_tol = np.where(s >= tol)[0]
    if len(above_tol) == 0:
        return range_space.empty(), np.array([]), np.zeros((0, len(A)))
    last_above_tol = above_tol[-1]

    errs = np.concatenate((np.cumsum(s[::-1] ** 2)[::-1], [0.]))
    below_err = np.where(errs <= l2_err**2)[0]
    first_below_err = below_err[0]

    selected_modes = min(first_below_err, last_above_tol + 1)
    if modes is not None:
        selected_modes = min(selected_modes, modes)

    U2 = U2[:, :selected_modes]
    s = s[:selected_modes]
    Vh = Vh[:selected_modes]

    logger.info(f'Computing left singular vectors ({selected_modes} modes) ...')
    U = Q.lincomb(U2.T)

    return U, s, Vh
